from typing import List, Tuple

import numpy as np
import tensorflow as tf  # type: ignore
from numpy.typing import NDArray

from code_loader.helpers.detection.utils import true_coords_labels, ciou, xywh_to_xyxy_format
from code_loader.helpers.detection.yolo.utils import match


class YoloLoss:
    """
    num classes - the number of classes to detect
    default_boxes - the anchors at all heads
    overlap_thresh - the threshold of IOU overwhich a match is positive
    neg_pos - the ratio of negative:positive samples
    background_label - should be the last idx
    """

    def __init__(self, num_classes: int, default_boxes: List[NDArray[np.int32]],
                 overlap_thresh: float, background_label: int,
                 from_logits: bool = True):
        self.background_label = background_label
        self.default_boxes = [tf.convert_to_tensor(box_arr) for box_arr in default_boxes]
        self.num_classes = num_classes
        self.threshold = overlap_thresh
        self.variance = (1, 1)
        self.ce = tf.keras.losses.CategoricalCrossentropy(from_logits=False, reduction='none')
        self.from_logits = from_logits
        # self.bce = tf.keras.losses.BinaryCrossentropy()
        self.weights = [4.0, 1.0, 0.4]  # Following yolov7 weights
        self.obj_w = 0.7
        self.cls_w = 0.00375
        self.box_w = 0.05

    def __call__(self, y_true: tf.Tensor, y_pred: Tuple[List[tf.Tensor], List[tf.Tensor]]) -> \
            Tuple[List[tf.Tensor], List[tf.Tensor], List[tf.Tensor]]:

        """
        Computes
        :param y_true:  Tensor, shape: (batch_size, MAX_BOXES_PER_IMAGE, 5(x, y, w, h, class_index))
        :param y_pred:  Tuple, (loc, conf) loc:
        :return: l_loss a list of NUM_FEATURES, each item a tensor with BATCH_SIZE legth
        :return: c_loss a list of NUM_FEATURES, each item a tensor with BATCH_SIZE legth
        """
        loc_data, conf_data = y_pred
        num = y_true.shape[0]
        l_losses = []
        c_losses = []
        o_losses = []
        for i in range(len(self.default_boxes)):
            default_box_layer = self.default_boxes[i]
            loc_data_layer = loc_data[i]
            conf_data_layer = conf_data[i][..., 1:]  # remove the "object confidence"
            priors = default_box_layer[:loc_data_layer.shape[1], :]
            # GT boxes
            loc_t = []
            conf_t = []
            priors = tf.cast(priors, dtype=tf.float32)  #
            y_true = tf.cast(y_true, dtype=tf.float32)
            for idx in range(num):
                truths, labels = true_coords_labels(idx, y_true, self.background_label)  #
                loc, conf = match(
                    threshold=self.threshold, truths=truths, priors=priors,
                    labels=labels,
                    background_label=self.background_label)  # encoded_gt_loc, label_pred, ciou between gt and pred
                loc_t.append(loc)
                conf_t.append(conf)
            loc_t_tensor: tf.Tensor = tf.stack(values=loc_t, axis=0)  # this is the location predictions (relative and logged)
            conf_t_tensor = tf.stack(values=conf_t, axis=0)  # these are the labels
            pos = conf_t_tensor != self.background_label
            num_pos = tf.reduce_sum(tf.cast(pos, dtype=tf.int32))

            # loss: Smooth L1 loss
            pos_idx = tf.expand_dims(pos, axis=-1)
            pos_idx = tf.broadcast_to(pos_idx, shape=loc_data_layer.shape)
            # apply per sample
            loss_b_list = []
            loss_c_list = []
            loss_o_list = []
            for j in range(num):
                loc_p = tf.boolean_mask(tensor=loc_data_layer[j, ...], mask=pos_idx[j, ...])
                object_p = conf_data[i][j, :, 0]  # object confidence
                targets_iou = tf.zeros_like(object_p)
                if self.from_logits:
                    sigmoid_obj = tf.sigmoid(object_p)
                else:
                    sigmoid_obj = object_p

                if loc_p.shape[0] and num_pos:  # GT exists
                    ious = ciou(xywh_to_xyxy_format(loc_data_layer[j, ...]),
                                xywh_to_xyxy_format(loc_t_tensor[j, ...]))
                    targets_iou = tf.where(pos_idx[j, :, 0], ious,
                                           tf.cast(tf.zeros_like(conf_t_tensor[j, ...]), tf.float32))
                    matched_iou = ious[pos_idx[j, :, 0]]
                    lbox = tf.reduce_mean(1. - matched_iou) * self.box_w
                    targets_iou = tf.convert_to_tensor(tf.maximum(targets_iou, 0).numpy())
                    # targets_iou = tf.stop_gradient(tf.maximum(targets_iou, 0))
                    if self.num_classes > 1:
                        if self.from_logits:
                            sig_pos_conf = tf.sigmoid(conf_data_layer[j])
                        else:
                            sig_pos_conf = conf_data_layer[j]

                        onehot_labels = tf.one_hot(conf_t_tensor[j], self.num_classes)
                        single_loss_cls = tf.reduce_mean(self.ce(onehot_labels, sig_pos_conf), axis=-1) * self.cls_w
                    else:
                        single_loss_cls = tf.constant(0, dtype=tf.float32)
                else:  # No GT
                    lbox = tf.zeros(1, dtype=tf.float32)
                    single_loss_cls = tf.constant(0, dtype=tf.float32)

                sigmoid_obj = tf.clip_by_value(sigmoid_obj, 1e-7, 1 - 1e-7)

                obj_loss = -(targets_iou * tf.math.log(sigmoid_obj) + (1 - targets_iou) * tf.math.log(1 - sigmoid_obj))
                mean_obj_loss = tf.reduce_mean(obj_loss) * self.obj_w * self.weights[i]

                loss_o_list.append(tf.expand_dims(mean_obj_loss, axis=0))
                loss_c_list.append(tf.expand_dims(single_loss_cls, axis=0))
                loss_b_list.append(tf.reshape(lbox, loss_c_list[-1].shape))

            loss_b_tensor = tf.stack(values=loss_b_list, axis=0)
            loss_o_tensor = tf.stack(values=loss_o_list, axis=0)
            loss_c_tensor = tf.stack(values=loss_c_list, axis=0)
            l_losses.append(loss_b_tensor)
            c_losses.append(loss_c_tensor)
            o_losses.append(loss_o_tensor)
        return l_losses, c_losses, o_losses



import sys
from types import TracebackType
from typing import List, Union

import numpy as np
import numpy.typing as npt

from code_loader.contract.datasetclasses import SectionCallableInterface, PreprocessResponse


def to_numpy_return_wrapper(encoder_function: SectionCallableInterface) -> SectionCallableInterface:
    def numpy_encoder_function(idx: int, samples: PreprocessResponse) -> npt.NDArray[np.float32]:
        result = encoder_function(idx, samples)
        numpy_result: npt.NDArray[np.float32] = np.array(result)
        return numpy_result

    return numpy_encoder_function


def get_root_traceback(exc_tb: TracebackType) -> TracebackType:
    return_traceback = exc_tb
    while return_traceback.tb_next is not None:
        return_traceback = return_traceback.tb_next
    return return_traceback


def get_root_exception_line_number() -> int:
    exc_tb = sys.exc_info()[2]
    root_exception_line_number = -1
    if exc_tb is not None:
        root_traceback = get_root_traceback(exc_tb)
        root_exception_line_number = root_traceback.tb_lineno
    return root_exception_line_number


def get_shape(result: Union[npt.NDArray[np.float32], str, float, int, bool]) -> List[int]:
    if not isinstance(result, np.ndarray):
        return [1]
    np_shape = result.shape
    # fix single result shape viewing
    if np_shape == ():
        np_shape = (1,)
    shape = list(np_shape)
    return shape


def rescale_min_max(image: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    image = image.astype('float32')
    image -= image.min()
    image /= (image.max() - image.min() + 1e-5)

    # rescale the values to range between 0 and 255
    image *= 255
    image = image.astype('uint8')

    return image





from typing import List

import numpy as np
import tensorflow as tf  # type: ignore
from numpy.typing import NDArray

from code_loader.helpers.detection.utils import xywh_to_xyxy_format
from code_loader.helpers.detection.yolo.utils import decode_bboxes


class Decoder:
    """At test time, Detect is the final layer of SSD.
    Consists of 4 steps:
    Bounding Boxes Decoding,
    Confidence Thresholding,
    Non-Max Suppression,
    Top-K Filtering.
    """

    def __init__(self, num_classes: int, background_label: int, top_k: int, conf_thresh: float, nms_thresh: float):
        self.num_classes = num_classes
        self.background_label = background_label
        self.top_k = top_k
        # Parameters used in nms.
        self.conf_thresh = conf_thresh
        self.nms_thresh = nms_thresh

    def __call__(self, loc_data: List[tf.Tensor], conf_data: List[tf.Tensor], prior_data: List[NDArray[np.float32]],
                 from_logits: bool = True, decoded: bool = False) -> List[NDArray[np.float32]]:
        """
        Args:
            loc_data: (tensor) Location preds from loc layers
                Shape: [batch, num_priors*4]
            conf_data: (tensor) Shape: Confidence preds from confidence layers
                Shape: [batch*num_priors, num_classes]
            prior_data: (tensor) Prior boxes and variances? from priorbox layers
                Shape: [1, num_priors, 4]
        """
        # loc_data: (batch_size, num_priors, 4)
        # conf_data: (batch_size, num_priors, num_classes)
        MAX_RETURNS = 20
        MAX_CANDIDATES_PER_LAYER = 20
        classes_num = conf_data[0].shape[-1]-1
        conf_preds = [tf.transpose(a=layer_conf, perm=[0, 2, 1]) for layer_conf in
                      conf_data]  # (batch_size, num_classes, num_priors)
        outputs = []
        for i in range(tf.shape(loc_data[0])[0]):
            loc = [loc_e[i, ...] for loc_e in loc_data]
            conf = [conf_e[i, ...] for conf_e in conf_preds]
            if from_logits:
                conf = [tf.math.sigmoid(layer_conf) for layer_conf in conf]
            class_selections: List[List[tf.Tensor]] = [[] for j in range(classes_num)]
            for l_loc, l_conf, l_prior in zip(loc, conf, prior_data):
                object_conf = l_conf[0, ...]
                if classes_num > 1:
                    l_conf = l_conf[1:, ...]*object_conf
                else:
                    l_conf = object_conf[np.newaxis, ...]
                mask = object_conf > self.conf_thresh
                classes = tf.argmax(l_conf, axis=0)
                max_scores = tf.reduce_max(l_conf, axis=0)
                # mask = max_scores > self.conf_thresh
                non_zero_indices = tf.where(mask)[:, 0]
                if len(non_zero_indices) != 0:
                    scores_masked = max_scores[mask]
                    if len(scores_masked) > MAX_CANDIDATES_PER_LAYER:
                        best_scores, best_indices = tf.math.top_k(scores_masked, k=MAX_CANDIDATES_PER_LAYER)
                    else:
                        best_scores = scores_masked
                        best_indices = np.arange(len(scores_masked))
                    topk_indices = tf.gather(non_zero_indices, best_indices)
                    selected_loc = tf.gather(l_loc, topk_indices, 1)
                    selected_scores = best_scores
                    if not decoded:
                        selected_prior = tf.gather(l_prior, topk_indices, 1)
                        selected_decoded = decode_bboxes(selected_loc.numpy(), selected_prior.numpy())  # (num_priors, 4)  (xmin, ymin, xmax, ymax) - THIS WORKS
                    else:
                        selected_decoded = xywh_to_xyxy_format(selected_loc.numpy())
                    selected_classes = tf.gather(classes, topk_indices)
                    for k in range(len(selected_classes)):
                        class_selections[selected_classes[k]].append(
                            (selected_scores[k], *selected_decoded[k, :], selected_classes[k]))
            final_preds = []
            for i in range(classes_num):
                if len(class_selections[i]) > 0:
                    np_selection: NDArray[np.float32] = np.array(class_selections[i])
                    boxes = np_selection[:, 1:5]
                    scores = np_selection[:, 0]
                    selected_indices = tf.image.non_max_suppression(boxes=boxes,
                                                                    scores=scores,
                                                                    max_output_size=self.top_k,
                                                                    iou_threshold=self.nms_thresh)
                    final_preds.append(np_selection[selected_indices, :])
            if any(class_selections):
                # choose best MAX_RETURNS/#detected class
                chosen_list = []
                remainder_list = []
                max_per_class = max(int(MAX_RETURNS/len(final_preds)), 1)
                for l in range(len(final_preds)):
                    chosen_list.append(final_preds[l][:max_per_class, ...])
                    remains = final_preds[l][max_per_class:, ...]
                    if remains.shape[0] > 0:
                        remainder_list.append(remains)
                predictions: NDArray[np.float32] = np.concatenate(chosen_list, axis=0)
                predictions = predictions[:MAX_RETURNS, ...]                       # this selects top 20 objects
                matched_obj = predictions.shape[0]
                missing_amount = MAX_RETURNS-matched_obj
                if len(remainder_list) > 0 and predictions.shape[0] < MAX_RETURNS:
                    remainder_np: NDArray[np.float32] = np.concatenate(remainder_list, axis=0)
                    if remainder_np.shape[0] > missing_amount:
                        top_indices = np.argpartition(remainder_np[:, 0], -missing_amount)[-missing_amount:]
                        remainder_np = remainder_np[top_indices, ...]
                    predictions = np.concatenate([predictions, remainder_np], axis=0)
                outputs.append(predictions)
            else:
                outputs.append(np.empty(0, dtype=float))
        return outputs



from enum import Enum, IntEnum


class Metric(Enum):
    MeanSquaredError = 'MeanSquaredError'
    MeanSquaredLogarithmicError = 'MeanSquaredLogarithmicError'
    MeanAbsoluteError = 'MeanAbsoluteError'
    MeanAbsolutePercentageError = 'MeanAbsolutePercentageError'
    Accuracy = 'Accuracy'
    BinaryAccuracy = 'BinaryAccuracy'
    MeanIOU = 'MeanIOU'
    ConfusionMatrixClassification = 'ConfusionMatrixClassification'
    ConfusionMatrixBinaryClassification = 'ConfusionMatrixBinaryClassification'


class LeapDataType(Enum):
    Image = 'Image'
    Text = 'Text'
    Graph = 'Graph'
    HorizontalBar = 'HorizontalBar'
    ImageMask = 'ImageMask'
    TextMask = 'TextMask'
    ImageWithBBox = 'ImageWithBBox'


class DatasetMetadataType(Enum):
    float = "float"
    string = "string"
    int = "int"
    boolean = "boolean"


class DataStateType(Enum):
    training = "training"
    validation = "validation"
    test = "test"
    unlabeled = "unlabeled"


class DataStateEnum(IntEnum):
    training = 0
    validation = 1
    test = 2
    unlabeled = 3


class ConfusionMatrixValue(Enum):
    Positive = "Positive"
    Negative = "Negative"


# todo: handle test not run due to error in pre process and Add to TestingSectionEnum Enum didn't run
class TestingSectionEnum(Enum):
    Warnings = "Warnings"
    Errors = "Errors"



import math
from typing import Tuple
from typing import Union

import numpy as np
import tensorflow as tf  # type: ignore
from numpy.typing import NDArray


def xyxy_to_xywh_format(boxes: Union[NDArray[np.float32], tf.Tensor]) -> Union[NDArray[np.float32], tf.Tensor]:
    """
    This gets bb in a [X,Y,W,H] format and transforms them into an [Xmin, Ymin, Xmax, Ymax] format
    :param boxes: [Num_boxes, 4] of type ndarray or tensor
    :return:
    """
    min_xy = (boxes[..., :2] + boxes[..., 2:]) / 2
    max_xy = (boxes[..., 2:] - boxes[..., :2])
    if isinstance(boxes, tf.Tensor):
        result = tf.concat([min_xy, max_xy], -1)
    else:
        result = np.concatenate([min_xy, max_xy], -1)
    return result


def xywh_to_xyxy_format(boxes: Union[NDArray[np.float32], tf.Tensor]) -> Union[NDArray[np.float32], tf.Tensor]:
    """
    This gets bb in a [X,Y,W,H] format and transforms them into an [Xmin, Ymin, Xmax, Ymax] format
    :param boxes: [Num_boxes, 4] of type ndarray or tensor
    :return:
    """
    min_xy = boxes[..., :2] - boxes[..., 2:] / 2
    max_xy = boxes[..., :2] + boxes[..., 2:] / 2
    if isinstance(boxes, tf.Tensor):
        result = tf.concat([min_xy, max_xy], -1)
    else:
        result = np.concatenate([min_xy, max_xy], -1)
    return result


def intersect(box_a: tf.Tensor, box_b: tf.Tensor) -> tf.Tensor:
    """

    :param box_a: Tensor, shape: (A, 4)
    :param box_b: Tensor, shape: (B, 4)
    :return: intersetction of box_a and box_b shape: (A, B)
    """
    min_xy = tf.math.minimum(tf.expand_dims(box_a[:, 2:], axis=1),
                             tf.expand_dims(box_b[:, 2:], axis=0))  # (right_bottom)
    max_xy = tf.math.maximum(tf.expand_dims(box_a[:, :2], axis=1),
                             tf.expand_dims(box_b[:, :2], axis=0))  # (left_top)
    inter = tf.math.maximum(min_xy - max_xy, 0)
    return inter[:, :, 0] * inter[:, :, 1]


def jaccard(box_a: tf.Tensor, box_b: tf.Tensor) -> tf.Tensor:
    """
    computes the IOU scores between two set of bb (box_a, box_b)
    :param box_a: Tensor, GT bounding boxes, shape: (a, 4)
    :param box_b: Tensor, other bounding boxes, shape: (b, 4)
    :return: Tensor, shape: (a, b)
    """
    inter = intersect(box_a, box_b)
    area_a = (box_a[:, 2] - box_a[:, 0]) * (box_a[:, 3] - box_a[:, 1])
    area_a = tf.expand_dims(area_a, axis=1)
    area_b = (box_b[:, 2] - box_b[:, 0]) * (box_b[:, 3] - box_b[:, 1])
    area_b = tf.expand_dims(area_b, axis=0)
    union = area_a + area_b - inter
    ratio = inter / union
    non_negativ_ratio = tf.where(inter > 0, x=ratio, y=0)
    return non_negativ_ratio


def true_coords_labels(idx: int, y_true: tf.Tensor, background_label: int) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    y_true shape  Tensor, shape: (batch_size, MAX_BOXES_PER_IMAGE, 5) (5 last channels are (xmin, ymin, xmax, ymax, class_index))
    This removes the background_class class from the ground_truth_labels
    :param idx:
    :param y_true:
    :return:
    """
    y_true = y_true[idx]
    mask = y_true[:, -1] != background_label  # class index should be greater than 0
    masked_true = tf.boolean_mask(y_true, mask)
    true_coords = masked_true[:, :-1]
    true_labels = masked_true[:, -1]
    return true_coords, true_labels


def ciou(box_a: tf.Tensor, box_b: tf.Tensor) -> tf.Tensor:
    """
    computes the IOU scores between two set of bb (box_a, box_b)
    :param box_a: Tensor, GT bounding boxes, shape: (a, 4)
    :param box_b: Tensor, other bounding boxes, shape: (a, 4)
    :return: Tensor, shape: a
    """
    eps = 1e-7
    # Intersection area
    inter = tf.maximum((tf.minimum(box_a[:, 2], box_b[:, 2]) - tf.maximum(box_a[:, 0], box_b[:, 0])), 0) * \
            tf.maximum((tf.minimum(box_a[:, 3], box_b[:, 3]) - tf.maximum(box_a[:, 1], box_b[:, 1])), 0)
    w1, h1 = box_a[:, 2] - box_a[:, 0], box_a[:, 3] - box_a[:, 1] + eps
    w2, h2 = box_b[:, 2] - box_b[:, 0], box_b[:, 3] - box_b[:, 1] + eps
    area_a = w1 * h1
    area_b = w2 * h2
    union = area_a + area_b - inter + eps
    iou = inter / union
    cw = tf.maximum(box_a[:, 2], box_b[:, 2]) - tf.minimum(box_a[:, 0],
                                                           box_b[:, 0])  # convex (smallest enclosing box) width
    ch = tf.maximum(box_a[:, 3], box_b[:, 3]) - tf.minimum(box_a[:, 1], box_b[:, 1])  # convex height
    c2 = cw ** 2 + ch ** 2 + eps  # convex diagonal squared
    rho2 = ((box_b[:, 0] + box_b[:, 2] - box_a[:, 0] - box_a[:, 2]) ** 2 +
            (box_b[:, 1] + box_b[:, 3] - box_a[:, 1] - box_a[:, 3]) ** 2) / 4  # center distance squared
    v = (4 / math.pi ** 2) * tf.pow(tf.atan(w2 / (h2 + eps)) - tf.atan(w1 / (h1 + eps)), 2)
    alpha = tf.stop_gradient(v / (v - iou + (1 + eps)))
    return iou - (rho2 / c2 + v * alpha)  # CIoU


def match(threshold: float, truths: tf.Tensor, priors: tf.Tensor, labels: tf.Tensor, background_label: int) \
        -> Tuple[tf.Tensor, tf.Tensor]:
    """
    Matches between the GT and the anchors
    :param threshold:
    :param truths: (N_truths,4) - (X,Y,W,H)
    :param priors: (N_priors,4) - (X,Y,W,H)
    :param labels: (N_truths) GT class
    :param background_label: int - the background label
    :return: loc (Tensor) [Npriors, 4], pred_label (Tensor) [Npriors]
    """
    # compute jaccard and best prior overlap and truth overlap
    overlaps = jaccard(xywh_to_xyxy_format(truths), xywh_to_xyxy_format(priors))  # (N_TRUTHS, N_PRIORS)
    best_prior_idx = tf.math.argmax(overlaps, axis=1)  # (NTRUTHS,)
    best_truth_overlap = tf.math.reduce_max(overlaps, axis=0, keepdims=True)  # (1, N_PRIORS)
    best_truth_idx = tf.math.argmax(overlaps, axis=0)  # (N_PRIORS,)
    # rates priors by GT overlap
    tf.squeeze(best_truth_overlap)
    tf.expand_dims(best_prior_idx, axis=1)
    tf.fill(dims=[tf.shape(best_prior_idx)[0]], value=2.0)
    best_truth_overlap = tf.tensor_scatter_nd_update(tensor=tf.squeeze(best_truth_overlap),
                                                     indices=tf.expand_dims(best_prior_idx, axis=1),
                                                     updates=tf.fill(dims=[tf.shape(best_prior_idx)[0]],
                                                                     value=2.0))  # (N_PRIORS)
    best_truth_overlap = tf.expand_dims(best_truth_overlap, axis=0)
    tf.expand_dims(best_prior_idx, axis=1)
    tf.range(start=0, limit=tf.shape(best_prior_idx)[0], delta=1,
             dtype=tf.int64)
    # For every PRIOR what is the best GT IDX
    best_truth_idx = tf.tensor_scatter_nd_update(tensor=best_truth_idx,
                                                 indices=tf.expand_dims(best_prior_idx, axis=1),
                                                 updates=tf.range(start=0, limit=tf.shape(best_prior_idx)[0], delta=1,
                                                                  dtype=tf.int64))
    # FOR EACH GT, replace the value of the best fitting prior with the GT INDEX
    # THIS RATES ALL GT ACCORDING TO WHICH RESULT IN HIGHEST JACACRD
    matches = tf.gather(params=truths, indices=best_truth_idx)  # GT for each PRIOR (N_PRIOR, 4)
    pred_label = tf.gather(params=labels, indices=best_truth_idx)  # THIS IS THE BEST LABELS
    pred_label = tf.where(condition=best_truth_overlap < threshold, x=background_label, y=tf.cast(pred_label,
                                                                                                  tf.int32))  # eliminates low threshold
    pred_label = tf.squeeze(pred_label)  # (Nprior)
    return matches, pred_label



from typing import Tuple, List

import numpy as np
from numpy.typing import NDArray

BoxSizeType = Tuple[Tuple[Tuple[float, float], ...], ...]


class Grid:
    def __init__(self, image_size: Tuple[int, int], feature_maps: Tuple[Tuple[int, int], ...], box_sizes: BoxSizeType,
                 strides: Tuple[int, ...], offset: int):
        self.image_size = image_size
        self.feature_maps = feature_maps
        self.box_sizes = box_sizes
        self.strides = strides
        self.offset = offset
        self.anchors = self.generate_cell_anchors()

    def generate_cell_anchors(self) -> NDArray[np.float32]:
        """
        This returns anchors, located at (0,0) sized according to to box_sizes.
        :return: np.ndarray of cell_anchors  (len(FEATURE_MAPS), number of anchors, 4) 4: X,Y,W,H
        """
        layer_anchors: List[NDArray[np.float32]] = []
        for layer_box_sizes in self.box_sizes:
            anchors = []
            for box_size in layer_box_sizes:
                x0, y0, w, h = 0., 0., box_size[0], box_size[1]
                anchors.append([x0, y0, w, h])
            layer_anchors.append(np.array(anchors))
        return np.stack(layer_anchors)

    def _create_grid_offsets(self, size: Tuple[int, int], stride: int) -> \
            Tuple[NDArray[np.float32], NDArray[np.float32]]:
        grid_height, grid_width = size
        shifts_x: NDArray[np.float32] = np.arange(- self.offset * stride, (grid_width - self.offset) * stride,
                                                  step=stride, dtype=np.float32)
        shifts_y: NDArray[np.float32] = np.arange(- self.offset * stride, (grid_height - self.offset) * stride,
                                                  step=stride,
                                                  dtype=np.float32)
        shift_x, shift_y = np.meshgrid(shifts_x, shifts_y)
        shift_x = shift_x.reshape(-1)
        shift_y = shift_y.reshape(-1)
        return shift_x, shift_y

    def generate_anchors(self) -> List[NDArray[np.float32]]:
        """
        Returns:
            list[Tensor]: #featuremap tensors, each is (#locations x #cell_anchors) x 4
        """
        anchors = []
        buffers = self.anchors
        grid_sizes = self.feature_maps
        for size, stride, base_anchors in zip(grid_sizes, self.strides, buffers):
            shift_x, shift_y = self._create_grid_offsets(size, stride)
            shifts: NDArray[np.float32] = np.stack((shift_x, shift_y, np.zeros_like(shift_x), np.zeros_like(shift_y)),
                                                   axis=1)
            absolute_anchors = (shifts.reshape((-1, 1, 4)) + base_anchors.reshape((1, -1, 4))).reshape(-1, 4)
            normalized_anchors = absolute_anchors / np.array([self.image_size[1], self.image_size[0],
                                                              self.image_size[1], self.image_size[0]])
            normalized_anchors = np.swapaxes(
                np.swapaxes(normalized_anchors.reshape((*size, len(base_anchors), 4)), 1, 2), 0, 1).reshape(-1, 4)
            anchors.append(normalized_anchors)
        return anchors



from typing import Tuple, List

import numpy as np
import tensorflow as tf  # type: ignore
from numpy.typing import NDArray

from code_loader.helpers.detection.utils import xywh_to_xyxy_format, jaccard

DEFAULT_FEATURE_MAPS = ((80, 80), (40, 40), (20, 20))
DEFAULT_BOX_SIZES = (((10, 13), (16, 30), (33, 23)),
                     ((30, 61), (62, 45), (59, 119)),
                     ((116, 90), (156, 198), (373, 326)))  # tiny fd


def decode_bboxes(loc_pred: tf.Tensor, priors: tf.Tensor, variances: int = 1) -> NDArray[np.float32]:
    """Decode locations from predictions using priors to undo
    the encoding we did for offset regression at train time.
    Args:
        loc_pred (tensor): location predictions for loc layers,
            Shape: [num_priors,4]
        priors (tensor): Prior boxes in center-offset form.
            Shape: [num_priors, 4].
        variances: (list[float]) Variances of priorboxes
    Return:
        decoded bounding box predictions as follows:
        x_pred = x_anchor + x_anchor*x_delta_pred
        y_pred = y_anchor + y_anchor*y_delta_pred
        w_pred = exp(w_delta_pred*var)*w_anchor
        h_pred = exp(h_delta_pred*var)*h_anchor
    """
    # MAX_CLIP = 4.135166556742356
    # log_preds = tf.clip_by_value(loc_pred[:, 2:], clip_value_max=MAX_CLIP, clip_value_min=-np.inf)
    boxes = tf.concat([
        priors[:, :2] + loc_pred[:, :2],
        priors[:, 2:] * (loc_pred[:, 2:] * variances) ** 2
    ], 1)
    return xywh_to_xyxy_format(boxes)


def encode_bboxes(matched: tf.Tensor, priors: tf.Tensor, variances: Tuple[int, int]) -> tf.Tensor:
    """
    This encodes a [X,Y,W,H] GT into a list of matches priors s.t.

    :param matched: Tensor - matched GT for each prior [N_PRIORS,4]
    :param priors: Tensor - Priors [N_priors,4]
    :param variances: Variances used
    :return:
        encoded bounding box gt Tensor
    """
    g_cxcy = (matched[:, :2] - priors[:, :2]) / (variances[0])
    g_wh = tf.math.sqrt((matched[:, 2:]) / priors[:, 2:]) / variances[1]
    return tf.concat([g_cxcy, g_wh], 1)


def scale_loc_prediction(loc_pred: List[tf.Tensor], decoded: bool = False, image_size: float = 640.,
                         strides: Tuple[int, int, int] = (8, 16, 32)) -> \
        List[tf.Tensor]:
    new_loc_pred = [None] * len(loc_pred)
    if decoded:
        new_loc_pred = [loc / image_size for loc in loc_pred]
    else:
        for i in range(len(loc_pred)):
            new_loc_pred[i] = tf.concat([(strides[i] * (2 * tf.sigmoid(loc_pred[i][..., :2]) - 0.5)) / image_size,
                                         2 * tf.sigmoid(loc_pred[i][..., 2:])], axis=-1)
    return new_loc_pred


def reshape_output_list(keras_output: tf.Tensor, image_size: int, priors: int = 3,
                        feature_maps: Tuple[Tuple[int, int], ...] = ((80, 80), (40, 40), (20, 20)),
                        decoded: bool = False) -> \
        Tuple[List[tf.Tensor], List[tf.Tensor]]:
    """
    reshape the mode's output to two lists sized [NUM_FEATURES] following detectron2 convention.
    class_list item: (BATCH_SIZE, NUM_ANCHORS, CLASSES)
    loc_list item:  (BATCH_SIZE, NUM_ANCHORS, 4)
    """
    num_features = len(feature_maps)
    j = 0
    loc_pred_list = []
    class_pred_list = []
    for k in range(num_features):
        # add classes prediction
        num_elements = feature_maps[k][0] * feature_maps[k][1] * priors
        loc_pred_list.append(keras_output[:, j:j + num_elements, :4])
        class_pred_list.append(keras_output[:, j:j + num_elements, 4:])
        j += num_elements
    loc_pred_list = scale_loc_prediction(loc_pred_list, decoded, image_size=image_size)
    return class_pred_list, loc_pred_list


def match(threshold: float, truths: tf.Tensor, priors: tf.Tensor
          , labels: tf.Tensor, background_label: int) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    Matches between the GT and the anchors
    :param threshold:
    :param truths: (N_truths,4) - (X,Y,W,H)
    :param priors: (N_priors,4) - (X,Y,W,H)
    :param variances:
    :param labels: (N_truths) GT class
    :param background_label: int - the background label
    :return: loc (Tensor) [Npriors, 4], pred_label (Tensor) [Npriors]
    """
    # compute jaccard and best prior overlap and truth overlap
    if truths.shape[0] == 0:  # NO GT
        matches = tf.zeros_like(priors)
        return matches, tf.ones(priors.shape[0], dtype=tf.int32) * background_label
    overlaps = jaccard(xywh_to_xyxy_format(truths), xywh_to_xyxy_format(priors))  # (N_TRUTHS, N_PRIORS)
    best_prior_idx = tf.math.argmax(overlaps, axis=1)  # (NTRUTHS,)
    best_truth_overlap = tf.math.reduce_max(overlaps, axis=0, keepdims=True)  # (1, N_PRIORS)
    best_truth_idx = tf.math.argmax(overlaps, axis=0)  # (N_PRIORS,)
    best_truth_idx = tf.tensor_scatter_nd_update(tensor=best_truth_idx,
                                                 indices=tf.expand_dims(best_prior_idx, axis=1),
                                                 updates=tf.range(start=0, limit=tf.shape(best_prior_idx)[0], delta=1,
                                                                  dtype=tf.int64))
    # FOR EACH GT, replace the value of the best fitting prior with the GT INDEX
    # THIS RATES ALL GT ACCORDING TO WHICH RESULT IN HIGHEST JACACRD
    matches = tf.gather(params=truths, indices=best_truth_idx)  # GT for each PRIOR (N_PRIOR, 4)
    pred_label = tf.gather(params=labels, indices=best_truth_idx)  # THIS IS THE BEST LABELS
    pred_label = tf.where(condition=best_truth_overlap < threshold, x=background_label, y=tf.cast(pred_label,
                                                                                                  tf.int32))  # eliminates low threshold
    pred_label = tf.squeeze(pred_label)  # (Nprior)
    return matches, pred_label  # decoded_gt, decoded_pred, ignore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['code_loader',
 'code_loader.contract',
 'code_loader.helpers',
 'code_loader.helpers.detection',
 'code_loader.helpers.detection.yolo',
 'code_loader.leap_binder',
 'code_loader.metrics',
 'code_loader.visualizers']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.22.3,<2.0.0', 'typeguard>=2.13.3,<3.0.0']

extras_require = \
{':platform_machine == "arm64"': ['tensorflow-macos>=2.11.0,<3.0.0'],
 ':platform_machine == "x86_64"': ['tensorflow>=2.11.0,<3.0.0']}

setup_kwargs = {
    'name': 'code-loader',
    'version': '0.2.48',
    'description': '',
    'long_description': '# tensorleap code loader\nUsed to load user code to tensorleap \n',
    'author': 'dorhar',
    'author_email': 'doron.harnoy@tensorleap.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tensorleap/code-loader',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
